// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include "server.h"
#include <math.h>
#include <gsl/gsl_multimin.h>

void fit_read_config(struct fitvars *fitconfig)
{
char name[1000];
FILE *config;
config = fopen("./fit.inp","r");
if (config == NULL)
{
	printf("no file %s\n",name);
	exit(0);
}

fscanf(config,"%s",name);
fscanf(config,"%lf",&(fitconfig->simplexmul));

fscanf(config,"%s",name);
fscanf(config,"%d",&(fitconfig->simplexreset));

fscanf(config,"%s",name);
fscanf(config,"%d",&(fitconfig->dolightfit));

fscanf(config,"%s",name);
fscanf(config,"%d",&(fitconfig->fit_light_ce));

fscanf(config,"%s",name);
fscanf(config,"%d",&(fitconfig->fit_rbi));

fscanf(config,"%s",name);
fscanf(config,"%d",&(fitconfig->fit_time_dark));

fscanf(config,"%s",name);
fscanf(config,"%d",&(fitconfig->fit_time_light));

fscanf(config,"%s",name);
fscanf(config,"%d",&(fitconfig->tpcmax));

fscanf(config,"%s",name);
fscanf(config,"%d",&(fitconfig->dotimece));

fscanf(config,"%s",name);
fscanf(config,"%d",&(fitconfig->dotimecelight));

fscanf(config,"%s",name);
fscanf(config,"%d",&(fitconfig->fit_dark_jv));

fscanf(config,"%s",name);
fscanf(config,"%d",&(fitconfig->fit_dark_ce_ss));

fscanf(config,"%s",name);
fscanf(config,"%lf",&(fitconfig->tpc_dark_error_mull));

fscanf(config,"%s",name);
fscanf(config,"%lf",&(fitconfig->tpc_light_error_mull));

fscanf(config,"%s",name);
fscanf(config,"%lf",&(fitconfig->ivdark_fitmul));

fscanf(config,"%s",name);
fscanf(config,"%lf",&(fitconfig->ivlight_fitmul));

fscanf(config,"%s",name);
fscanf(config,"%lf",&(fitconfig->chargedark_fitmul));

fscanf(config,"%s",name);
fscanf(config,"%lf",&(fitconfig->chargelight_fitmul));

fscanf(config,"%s",name);
fscanf(config,"%lf",&(fitconfig->rbi_fitmul));

fscanf(config,"%s",name);
fscanf(config,"%lf",&(fitconfig->banderror_fitmul));

fscanf(config,"%s",name);
fscanf(config,"%lf",&(fitconfig->muerror_fitmul));

fscanf(config,"%s",name);
fscanf(config,"%d",&fitconfig->fitvars);

int i;
int items=0;
char file[20];
int pos;
int dofit;
for (i=0;i<fitconfig->fitvars;i++)
{
fscanf(config,"%s",name);
fscanf(config,"%s",file);
fscanf(config,"%d",&pos);
fscanf(config,"%d",&dofit);

if (dofit==TRUE)
{
strcpy(fitconfig->fit_file[items],file);
fitconfig->fit_pos[items]=pos;
items++;
}

printf("%d %s %s %d %d\n",i,name,file,pos,dofit);
}
fitconfig->fitvars=items;
fclose(config);
}
