// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include <math.h>

double newton_sim_tpv_wire(struct device *in)
{
double step=0.001;
double e0;
double e1;
double i0;
double i1;
double deriv;
solve_all(in);
i0=get_I(in);
//getchar();
e0=fabs(i0+in->Vapplied/in->Rshunt+wire_sim(in->Vapplied));
in->Vapplied+=step;
solve_all(in);
i1=get_I(in);
e1=fabs(i1+in->Vapplied/in->Rshunt+wire_sim(in->Vapplied));
deriv=(e1-e0)/step;
step=-e1/deriv;
in->Vapplied+=step;
int count=0;
int max=100;
do
{
e0=e1;
solve_all(in);
i1=get_I(in);
e1=fabs(i1+in->Vapplied/in->Rshunt+wire_sim(in->Vapplied));
deriv=(e1-e0)/step;
step=-e1/deriv;
double clamp=0.01;
if (e1<clamp) clamp=e1/10.0;
step=step/(1.0+fabs(step/clamp));
in->Vapplied+=step;
printf("Tpv error = %e %e\n",in->Vapplied,e1);
if (count>max) break;
count++;
}while(e1>1e-8);
double cur=wire_sim(in->Vapplied);
printf("result = %e %e %e %e\n",i1,in->Vapplied/in->Rshunt,cur,in->Vapplied/cur);
return 0.0;
}

double sim_tpv_simple(struct device *in)
{
double step=0.001;
double e0;
double e1;
double i0;
double i1;
double deriv;
double C=24e-12*10.0;
double L=0.059e-6*10.0;
solve_all(in);
i0=get_I(in);
//getchar();
e0=fabs(i0+in->Vapplied/in->Rshunt+C*(in->Vapplied-in->Vapplied_last)/in->dt-L*C*(i0-in->Ilast)/in->dt);
in->Vapplied+=step;
solve_all(in);
i1=get_I(in);
e1=fabs(i1+in->Vapplied/in->Rshunt+C*(in->Vapplied-in->Vapplied_last)/in->dt-L*C*(i1-in->Ilast)/in->dt);
deriv=(e1-e0)/step;
step=-e1/deriv;
in->Vapplied+=step;
int count=0;
int max=10;
do
{
e0=e1;
solve_all(in);
i1=get_I(in);
e1=fabs(i1+in->Vapplied/in->Rshunt+C*(in->Vapplied-in->Vapplied_last)/in->dt-L*C*(i1-in->Ilast)/in->dt);
deriv=(e1-e0)/step;
step=-e1/deriv;
//double clamp=0.01;
//if (e1<clamp) clamp=e1/100.0;
//step=step/(1.0+fabs(step/clamp));
in->Vapplied+=step;
printf("Tpv error = %e %e\n",in->Vapplied,e1);
if (count>max) break;
count++;
}while(e1>1e-6);
//printf("result = %e %e %e %e %e\n",i1,in->Vapplied/in->Rshunt,C*(in->Vapplied-in->Vapplied_last)/in->dt,i1+in->Vapplied/in->Rshunt+C*(in->Vapplied-in->Vapplied_last)/in->dt);
double ret=in->Vapplied-C*(i1-in->Ilast)/in->dt;
printf("\n\nreturning %e\n\n",ret);
return ret;
}

double sim_i(struct device *in,double wantedi)
{
double clamp=0.01;
double step=0.01;
double e0;
double e1;
double i0;
double i1;
double deriv;

solve_all(in);
i0=get_equiv_I(in);
e0=fabs(wantedi-i0);
in->Vapplied+=step;
solve_all(in);
i1=get_equiv_I(in);
e1=fabs(wantedi-i1);
deriv=(e1-e0)/step;
step=-e1/deriv;

in->Vapplied+=step;
int count=0;
int max=1000;
do
{
e0=e1;
solve_all(in);
i1=get_equiv_I(in);
e1=fabs(wantedi-i1);
deriv=(e1-e0)/step;
step=-e1/deriv;

//if (e1<clamp) clamp=e1/100.0;
step=step/(1.0+fabs(step/clamp));
in->Vapplied+=step;
//printf("I error = %d %e %e %e \n",count,in->Vapplied,e1,clamp);
if (count>max) break;
count++;
}while(e1>1e-6);

return in->Vapplied;
}

double sim_voc(struct device *in,double R,int prime)
{
double clamp=0.001;
double step=0.01;
double e0;
double e1;
double deriv;
//in->Vapplied=0.2;
solve_all(in);
e1=fabs(get_I(in)*R-in->Vapplied);
//printf("Vapplied=%e\n",in->Vapplied);
//getchar();
in->Vapplied+=step;
int count=0;
int max=10;
if (prime==TRUE) max=100;
do
{
e0=e1;
solve_all(in);
e1=fabs(get_I(in)*R-in->Vapplied);
printf("%d Vr=%e Vcell=%e I=%e Error=%e Jrbi=%e Jrbi=%e\n",count,get_I(in)*R,in->Vapplied,get_I_recomb(in),e1,get_J(in),get_J(in));
//getchar();
deriv=(e1-e0)/step;

step=-e1/deriv;
//if (e1<1.0) clamp=e1/100.0;
step=step/(1.0+fabs(step/clamp));
in->Vapplied+=step;
//printf("Applied = %e OverR = %e Error = %e %e \n",in->Vapplied,get_I(in)*in->Rshunt,e1,get_I(in));
if (count>max) break;
count++;
}while(e1>1e-4);
printf("Finished\n");
//getchar();
return in->Vapplied;
}

double sim_externalv(struct device *in,double wantedv)
{
double step=0.01;
double e0;
double e1;
double v0;
double v1;
double deriv;
double maxerror=1e-5;
double clamp=0.0;
//in->Vapplied=wantedv;
solve_all(in);
v0=get_equiv_V(in);
e0=fabs(wantedv-v0);
//if (e0<maxerror) return in->Vapplied;
in->Vapplied+=step;
solve_all(in);
v1=get_equiv_V(in);
e1=fabs(wantedv-v1);
//if (e1<maxerror) return in->Vapplied;
deriv=(e1-e0)/step;
step=-e1/deriv;
clamp=0.1;
step=step/(1.0+fabs(step/clamp));
in->Vapplied+=step;

int count=0;
int max=200;
do
{
e0=e1;
solve_all(in);
v1=get_equiv_V(in);
e1=fabs(wantedv-v1);
deriv=(e1-e0)/step;
step=-e1/deriv;
clamp=0.1;
step=step/(1.0+fabs(step/clamp));
in->Vapplied+=step;
#ifndef _fast_
printf("Applied = %e External = %e Error = %e %e %e\n",in->Vapplied,v1,e1,step,step/(1.0+fabs(step/clamp)));
#endif
if (count>max) break;
count++;
}while(e1>maxerror);

return in->Vapplied;
}


void sim_jv(struct device *in)
{
double V=in->Vstart;
double Vstop=in->Vstop;
double Vstep=in->Vstep;
FILE *iv;
int ittr=0;
double J;
double Pden;
int first=TRUE;
double Vlast;
double Jlast;
double Pdenlast;
double Vexternal;
struct istruct ivexternal;
inter_init(&ivexternal);

struct istruct charge;
inter_init(&charge);

struct istruct klist;
inter_init(&klist);

iv=fopena(in->outputpath,"./jv.dat","w");
fclose(iv);

iv=fopena(in->outputpath,"./avgmu.dat","w");
fclose(iv);

iv=fopena(in->outputpath,"./freetrapped.dat","w");
fclose(iv);


#ifndef mindump
iv=fopen("./k3d.dat","w");
fclose(iv);
#endif


struct map frame;
double srh_stop=0.8;
double srh_den=10.0*in->srh_bands/srh_stop;

frame_init(&frame,in->ymeshpoints,1000,in->ymesh[0],in->ymesh[in->ymeshpoints-1],-6.0,-4.0);

	do
	{

		in->Vapplied=V;



		solve_all(in);

		J=get_equiv_J(in);
		Vexternal=get_equiv_V(in);

		#ifndef mindump
		iv=fopen("./k3d.dat","a");
		fprintf(iv, "%le %le %le\n",Vexternal, in->Psun/1000.0,get_avg_k(in));
		fclose(iv);
		#endif


		frame_reset(&frame);
		frame_build(&frame,in);
		frame_dump(&frame,in);
		//getchar();

		if (ittr>0)
		{

			inter_append(&ivexternal,get_equiv_V(in),get_equiv_J(in));

		}

		ittr++;
		//iv=fopena(in->outputpath,"./freetrapped.dat","a");




		//fprintf(iv,"%e %e %e %e %e\n",V,get_total_n_trapped_charge(in),get_total_n_charge(in),get_total_p_trapped_charge(in),get_total_p_charge(in));
		//fclose(iv);


		inter_append(&charge,get_equiv_V(in),get_extracted_np(in));


			Pden=fabs(J*Vexternal);




			plot_now(in,"plot");


			if (first==FALSE)
			{

				//check if we have crossed 0V
				if ((Vlast<=0)&&(V>=0.0))
				{
					in->Jsc=Jlast+(J-Jlast)*(0-Vlast)/(V-Vlast);
					printf("Jsc = %e\n",in->Jsc);
				}

				if ((Jlast<=0)&&(J>=0.0))
				{
					in->Voc=Vlast+(V-Vlast)*(0-Jlast)/(J-Jlast);
					printf("Voc = %e\n",in->Voc);
					FILE *voc_file;
					voc_file=fopena(in->outputpath,"./voc.dat","w");
					fprintf(voc_file,"%e\n",in->Voc);
					fclose(voc_file);
				}

				if ((Pden>Pdenlast)&&(Vexternal>0.0)&&(J<0.0))
				{
					in->Pmax=Pden;
					in->Pmax_voltage=Vexternal;
				}

				if (Vexternal>Vstop) break;

			}

			#ifndef _fast_
			printf("V=%lf %lf current = %e mA (%e A/m^2) %le\n",V,Vexternal,get_I(in)/1e-3,J,get_J(in));
			#endif

			if (in->stop_start==TRUE) getchar();


			Jlast=J;
			Vlast=Vexternal;
			Pdenlast=Pden;
			first=FALSE;

			V+=Vstep;
			//dialog_set_progress ((in->Vstart+V)/(in->Vstop-in->Vstart));
			if ((Vstep>=0)&&(V>Vstop))
			{
				in->stop=TRUE;
				printf("--------------1\n");
			}

			if ((Vstep<0)&&(V<Vstop))
			{
				in->stop=TRUE;
				printf("--------------2\n");
			}
//printf("%e\n",Vstop);



		if (in->stop==TRUE)
		{
			break;

		}

	}while(1);

in->FF=fabs(in->Pmax/(in->Jsc*in->Voc));

char out[200];
sprintf(out,"Voc= %lf (V)\n",in->Voc);
printdebug( out);
sprintf(out,"Jsc= %lf (A/m^2)\n",in->Jsc);
printdebug( out);
sprintf(out,"Pmax= %lf (W/m^2)\n",in->Pmax);
printdebug( out);
sprintf(out,"Voltage to get Pmax= %lf (V)\n",in->Pmax_voltage);
printdebug( out);
sprintf(out,"FF= %lf\n",in->FF*100.0);
printdebug( out);

#ifndef mindump
iv=fopen("./ff.dat","w");
fprintf(iv,"%le\n",in->FF*100.0);
fclose(iv);
#endif

sprintf(out,"Efficiency= %lf percent\n",mod(in->Pmax/in->Psun)*100.0);
printdebug( out);

#ifndef mindump
FILE *outfile=fopen("./eff.dat","w");
fprintf(outfile,"%le",mod(in->Pmax/(in->Psun))*100.0);
fclose(outfile);
#endif

inter_save_a(&charge,in->outputpath,"charge.dat");
inter_free(&charge);

inter_save_a(&ivexternal,in->outputpath,"ivexternal.dat");
inter_free(&ivexternal);

inter_save_a(&klist,in->outputpath,"k.dat");
inter_free(&klist);
frame_free(&frame);
}

