// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include <math.h>

void free_srh_bands(struct device *in,double ** var)
{
	int i;
	for (i=0;i<in->ymeshpoints;i++)
	{
		free(var[i]);
	}

free(var);
}

void malloc_srh_bands(struct device *in,double *(** var))
{
	*var = (double **)malloc(in->ymeshpoints*sizeof(double*));

	int i;
	for (i=0;i<in->ymeshpoints;i++)
	{
		(*var)[i] = (double *)malloc(in->srh_bands*sizeof(double));
	}
}


void device_free(struct device *in)
{
	free(in->phi);
	free(in->Nad);
	free(in->n);
	free(in->p);
	free(in->dn);
	free(in->dp);
	free(in->Eg);
	free(in->Xi);
	free(in->Ev);
	free(in->Ec);
	free(in->mun);
	free(in->mup);
	free(in->ymesh);
	free(in->Fn);
	free(in->Fp);

	free(in->Nc);
	free(in->Nv);
	free(in->me);
	free(in->mh);
	free(in->G);
	free(in->Gex);
	free(in->Tl);
	free(in->Te);
	free(in->Th);
	free(in->R);
	free(in->ymatlayers);
	free(in->Fi);
	free(in->Jn);
	free(in->Jp);
	free(in->x);
	free(in->t);
	free(in->xp);
	free(in->tp);
	free(in->ex);
	free(in->Dex);
	free(in->Hex);
	free(in->epsilonr);

	free(in->back_x);
	free(in->back_xp);
	free(in->back_phi);
	free(in->back_xt);
	free(in->back_xpt);

	free(in->kf);
	free(in->kd);
	free(in->kr);
	free(in->a);
	free(in->Rbi);
	free(in->Rn);
	free(in->Rp);
	free(in->kl);
	free(in->ke);
	free(in->kh);
	free(in->Hl);
	free(in->He);
	free(in->Hh);
	free(in->Hasorb);
	free(in->nlast);
	free(in->plast);
	free(in->ntlast);
	free(in->ptlast);

	free(in->wn);
	free(in->wp);

	free(in->nt_all);
	free_srh_bands(in,in->nt);
	free_srh_bands(in,in->xt);
	free_srh_bands(in,in->dnt);
	free_srh_bands(in,in->srh_n_r1);
	free_srh_bands(in,in->srh_n_r2);
	free_srh_bands(in,in->srh_n_r3);
	free_srh_bands(in,in->srh_n_r4);
	free_srh_bands(in,in->dsrh_n_r1);
	free_srh_bands(in,in->dsrh_n_r2);
	free_srh_bands(in,in->dsrh_n_r3);
	free_srh_bands(in,in->dsrh_n_r4);
	free_srh_bands(in,in->Fnt);
	free_srh_bands(in,in->ntbinit);

	free(in->Nct);
	free(in->tt);
	free(in->dostype);
	free(in->Rbi_k);

	free(in->pt_all);
	free_srh_bands(in,in->pt);
	free_srh_bands(in,in->dpt);
	free_srh_bands(in,in->xpt);
	free_srh_bands(in,in->srh_p_r1);
	free_srh_bands(in,in->srh_p_r2);
	free_srh_bands(in,in->srh_p_r3);
	free_srh_bands(in,in->srh_p_r4);
	free_srh_bands(in,in->dsrh_p_r1);
	free_srh_bands(in,in->dsrh_p_r2);
	free_srh_bands(in,in->dsrh_p_r3);
	free_srh_bands(in,in->dsrh_p_r4);
	free_srh_bands(in,in->Fpt);
	free_srh_bands(in,in->ptbinit);

	free(in->Nvt);
	free(in->tpt);



	free(in->A);
	free(in->deltaE);

	free(in->nfinit);
	free(in->pfinit);
	free(in->ntinit);
	free(in->ptinit);

	//fclose(in->gnuplot);

	free(in->meshdata);
	int i;
	printf("mat number= %d\n",in->mat.number);
	for (i=0;i<in->mat.number;i++)
	{
		mat_free(&(in->mat.l[i]));
	}

	mat_free(&(in->lcontact));
	mat_free(&(in->rcontact));
	free(in->mat.l);
	free(in->capeff);
	solver_free();

	free(in->nrelax);
	free(in->ntrap_to_p);
	free(in->prelax);
	free(in->ptrap_to_n);

	printf("Solved %i Equations\n",in->odes);
}


void device_get_memory(struct device *in)
{
	in->odes=0;

	in->nfinit = malloc(in->ymeshpoints*sizeof(double));
	in->pfinit = malloc(in->ymeshpoints*sizeof(double));
	in->ntinit = malloc(in->ymeshpoints*sizeof(double));
	in->ptinit = malloc(in->ymeshpoints*sizeof(double));

	in->Hasorb = malloc(in->ymeshpoints*sizeof(double));
	in->phi = malloc(in->ymeshpoints*sizeof(double));
	in->Nad = malloc(in->ymeshpoints*sizeof(double));
	in->n = malloc(in->ymeshpoints*sizeof(double));
	in->p = malloc(in->ymeshpoints*sizeof(double));
	//malloc_srh_bands(in,&(in->n));
	//malloc_srh_bands(in,&(in->p));
	in->dn = malloc(in->ymeshpoints*sizeof(double));
	in->dp = malloc(in->ymeshpoints*sizeof(double));
	//malloc_srh_bands(in,&(in->dn));
	//malloc_srh_bands(in,&(in->dp));
	in->Eg = malloc(in->ymeshpoints*sizeof(double));
	in->Fn = malloc(in->ymeshpoints*sizeof(double));
	in->Fp = malloc(in->ymeshpoints*sizeof(double));
	//malloc_srh_bands(in,&(in->Fn));
	//malloc_srh_bands(in,&(in->Fp));

	in->Xi = malloc(in->ymeshpoints*sizeof(double));
	in->Ev = malloc(in->ymeshpoints*sizeof(double));
	in->Ec = malloc(in->ymeshpoints*sizeof(double));
	//malloc_srh_bands(in,&(in->mun));
	//malloc_srh_bands(in,&(in->mup));
	in->mun = malloc(in->ymeshpoints*sizeof(double));
	in->mup = malloc(in->ymeshpoints*sizeof(double));

	in->Nc = malloc(in->ymeshpoints*sizeof(double));
	in->Nv = malloc(in->ymeshpoints*sizeof(double));
	in->me = malloc(in->ymeshpoints*sizeof(double));
	in->mh = malloc(in->ymeshpoints*sizeof(double));
	in->G = malloc(in->ymeshpoints*sizeof(double));
	in->Gex = malloc(in->ymeshpoints*sizeof(double));
	in->Tl = malloc (in->ymeshpoints * sizeof(double));
	in->Te = malloc (in->ymeshpoints * sizeof(double));
	in->Th = malloc (in->ymeshpoints * sizeof(double));

	in->R = malloc (in->ymeshpoints * sizeof(double));
	in->Fi = malloc (in->ymeshpoints * sizeof(double));
	//malloc_srh_bands(in,&(in->Jn));
	//malloc_srh_bands(in,&(in->Jp));
	in->Jn = malloc (in->ymeshpoints * sizeof(double));
	in->Jp = malloc (in->ymeshpoints * sizeof(double));
	//malloc_srh_bands(in,&(in->x));
	//malloc_srh_bands(in,&(in->t));
	in->x = malloc (in->ymeshpoints * sizeof(double));
	in->t = malloc (in->ymeshpoints * sizeof(double));

	//malloc_srh_bands(in,&(in->xp));
	//malloc_srh_bands(in,&(in->tp));
	in->xp = malloc (in->ymeshpoints * sizeof(double));
	in->tp = malloc (in->ymeshpoints * sizeof(double));

	in->kf = malloc (in->ymeshpoints * sizeof(double));
	in->kd = malloc (in->ymeshpoints * sizeof(double));
	in->kr = malloc (in->ymeshpoints * sizeof(double));
	in->a = malloc (in->ymeshpoints * sizeof(double));
	in->Rbi = malloc (in->ymeshpoints * sizeof(double));
	in->Rn = malloc (in->ymeshpoints * sizeof(double));
	in->Rp = malloc (in->ymeshpoints * sizeof(double));

	in->back_x = malloc (in->ymeshpoints * sizeof(double));
	in->back_xp = malloc (in->ymeshpoints * sizeof(double));
	in->back_phi = malloc(in->ymeshpoints*sizeof(double));
	in->back_xt = malloc(in->ymeshpoints*sizeof(double));
	in->back_xpt = malloc(in->ymeshpoints*sizeof(double));


	in->ex = malloc (in->ymeshpoints * sizeof(double));
	in->Dex = malloc (in->ymeshpoints * sizeof(double));
	in->Hex = malloc (in->ymeshpoints * sizeof(double));
	in->epsilonr = malloc (in->ymeshpoints * sizeof(double));
	in->kl = malloc(in->ymeshpoints*sizeof(double));
	in->ke = malloc(in->ymeshpoints*sizeof(double));
	in->kh = malloc(in->ymeshpoints*sizeof(double));
	in->Hl = malloc(in->ymeshpoints*sizeof(double));
	in->He = malloc(in->ymeshpoints*sizeof(double));
	in->Hh = malloc(in->ymeshpoints*sizeof(double));

	in->nlast = malloc(in->ymeshpoints*sizeof(double));
	in->plast = malloc(in->ymeshpoints*sizeof(double));

	in->wn = malloc(in->ymeshpoints*sizeof(double));
	in->wp = malloc(in->ymeshpoints*sizeof(double));

	in->Nct = malloc(in->ymeshpoints*sizeof(double));
	in->nt_all = malloc(in->ymeshpoints*sizeof(double));
	malloc_srh_bands(in,&(in->nt));
	malloc_srh_bands(in,&(in->ntlast));
	malloc_srh_bands(in,&(in->xt));
	malloc_srh_bands(in,&(in->dnt));
	malloc_srh_bands(in,&(in->srh_n_r1));
	malloc_srh_bands(in,&(in->srh_n_r2));
	malloc_srh_bands(in,&(in->srh_n_r3));
	malloc_srh_bands(in,&(in->srh_n_r4));
	malloc_srh_bands(in,&(in->dsrh_n_r1));
	malloc_srh_bands(in,&(in->dsrh_n_r2));
	malloc_srh_bands(in,&(in->dsrh_n_r3));
	malloc_srh_bands(in,&(in->dsrh_n_r4));
	malloc_srh_bands(in,&(in->Fnt));
	malloc_srh_bands(in,&(in->ntbinit));
	in->tt = malloc(in->ymeshpoints*sizeof(double));
	in->dostype = malloc(in->ymeshpoints*sizeof(double));

	in->pt_all = malloc(in->ymeshpoints*sizeof(double));
	malloc_srh_bands(in,&(in->pt));
	malloc_srh_bands(in,&(in->ptlast));
	malloc_srh_bands(in,&(in->xpt));
	malloc_srh_bands(in,&(in->dpt));
	malloc_srh_bands(in,&(in->srh_p_r1));
	malloc_srh_bands(in,&(in->srh_p_r2));
	malloc_srh_bands(in,&(in->srh_p_r3));
	malloc_srh_bands(in,&(in->srh_p_r4));
	malloc_srh_bands(in,&(in->dsrh_p_r1));
	malloc_srh_bands(in,&(in->dsrh_p_r2));
	malloc_srh_bands(in,&(in->dsrh_p_r3));
	malloc_srh_bands(in,&(in->dsrh_p_r4));
	malloc_srh_bands(in,&(in->ptbinit));
	malloc_srh_bands(in,&(in->Fpt));

	in->Nvt = malloc(in->ymeshpoints*sizeof(double));
	in->tpt = malloc(in->ymeshpoints*sizeof(double));
	in->A = malloc(in->ymeshpoints*sizeof(double));
	in->deltaE = malloc(in->ymeshpoints*sizeof(double));
	in->capeff = malloc(in->ymeshpoints*sizeof(double));
	in->Rbi_k = malloc(in->ymeshpoints*sizeof(double));

	in->nrelax=malloc(in->ymeshpoints*sizeof(double));
	in->ntrap_to_p=malloc(in->ymeshpoints*sizeof(double));
	in->prelax=malloc(in->ymeshpoints*sizeof(double));
	in->ptrap_to_n=malloc(in->ymeshpoints*sizeof(double));

}
